@@ -112,7 +112,36 @@ getprogname (void)
 	if (p != NULL)
 		name = p + 1;
 #elif defined (HAVE_PROGRAM_INVOCATION_SHORT_NAME)
+#ifdef __linux__
+	name = program_invocation_name;
+	assert (name);
+	if (*name == '/') {
+		/*
+		 * Some programs pack command line arguments into argv[0].
+		 * Check if it is the case by reading /proc/self/exe and extract
+		 * the program name.
+		 *
+		 * Logic borrowed from:
+		 * <https://github.com/mesa3d/mesa/commit/759b94038987bb983398cd4b1d2cb1c8f79817a9>.
+		 */
+		static char *buf;
+
+		if (!buf)
+			buf = realpath ("/proc/self/exe", NULL);
+
+		if (buf && strncmp (buf, name, strlen (buf)) == 0)
+			/* Use the executable path if the prefix matches. */
+			name = strrchr (buf, '/') + 1;
+		else
+			/* Otherwise fall back to
+			 * program_invocation_short_name. */
+			name = program_invocation_short_name;
+	} else {
+		name = program_invocation_short_name;
+	}
+#else
 	name = program_invocation_short_name;
+#endif
 #elif defined (HAVE___PROGNAME)
 	name = __progname;
 #else
@@ -496,8 +525,8 @@ reallocarray (void *ptr,
 	      size_t nmemb,
 	      size_t size)
 {
-	assert (nmemb > 0 && size > 0);
-	if (SIZE_MAX / nmemb < size) {
+	assert (nmemb >= 0 && size >= 0);
+	if (nmemb != 0 && SIZE_MAX / nmemb < size) {
 		errno = ENOMEM;
 		return NULL;
 	}
@@ -816,7 +845,7 @@ getauxval (unsigned long type)
 		secure = __libc_enable_secure;
 
 #elif defined(HAVE_ISSETUGID) && \
-	!((defined __APPLE__ && defined __MACH__) || (defined __FREEBSD__))
+	!((defined __APPLE__ && defined __MACH__) || (defined __FreeBSD__))
 		secure = issetugid ();
 
 #elif defined(OS_UNIX)
@@ -909,9 +938,6 @@ int
 fdwalk (int (* cb) (void *data, int fd),
         void *data)
 {
-	struct dirent *de;
-	char *end;
-	DIR *dir;
 	int open_max;
 	long num;
 	int res = 0;
@@ -922,6 +948,10 @@ fdwalk (int (* cb) (void *data, int fd),
 #endif
 
 #ifdef __linux__
+	struct dirent *de;
+	char *end;
+	DIR *dir;
+
 	dir = opendir ("/proc/self/fd");
 	if (dir != NULL) {
 		while ((de = readdir (dir)) != NULL) {
