@@ -43,6 +43,7 @@
 #include "rpc-message.h"
 
 #include <assert.h>
+#include <errno.h>
 #include <string.h>
 
 #define ELEMS(x) (sizeof (x) / sizeof (x[0]))
@@ -114,6 +115,18 @@ p11_rpc_message_alloc_extra (p11_rpc_message *msg,
 	return (void *)(data + 1);
 }
 
+void *
+p11_rpc_message_alloc_extra_array (p11_rpc_message *msg,
+				   size_t nmemb,
+				   size_t size)
+{
+	if (nmemb != 0 && (SIZE_MAX - sizeof (void *)) / nmemb < size) {
+		errno = ENOMEM;
+		return NULL;
+	}
+	return p11_rpc_message_alloc_extra (msg, nmemb * size);
+}
+
 bool
 p11_rpc_message_prep (p11_rpc_message *msg,
                       int call_id,
@@ -744,7 +757,7 @@ p11_rpc_buffer_get_byte_array (p11_buffer *buf,
 		return false;
 	}
 
-	if (buf->len < len || *offset > buf->len - len) {
+	if (buf->len < len || off > buf->len - len) {
 		p11_buffer_fail (buf);
 		return false;
 	}
@@ -875,6 +888,8 @@ static p11_rpc_attribute_serializer p11_rpc_attribute_serializers[] = {
 	{ P11_RPC_VALUE_BYTE_ARRAY, p11_rpc_buffer_add_byte_array_value, p11_rpc_buffer_get_byte_array_value }
 };
 
+P11_STATIC_ASSERT(sizeof(CK_BYTE) <= sizeof(uint8_t));
+
 void
 p11_rpc_buffer_add_byte_value (p11_buffer *buffer,
 			       const void *value,
@@ -890,12 +905,6 @@ p11_rpc_buffer_add_byte_value (p11_buffer *buffer,
 	if (value)
 		memcpy (&byte_value, value, value_length);
 
-	/* Check if byte_value can be converted to uint8_t. */
-	if (byte_value > UINT8_MAX) {
-		p11_buffer_fail (buffer);
-		return;
-	}
-
 	p11_rpc_buffer_add_byte (buffer, byte_value);
 }
 
@@ -953,9 +962,7 @@ p11_rpc_buffer_add_mechanism_type_array_value (p11_buffer *buffer,
 					       const void *value,
 					       CK_ULONG value_length)
 {
-	const CK_MECHANISM_TYPE *mechs = value;
 	size_t count = value_length / sizeof (CK_MECHANISM_TYPE);
-	size_t i;
 
 	/* Check if count can be converted to uint32_t. */
 	if (count > UINT32_MAX) {
@@ -966,12 +973,17 @@ p11_rpc_buffer_add_mechanism_type_array_value (p11_buffer *buffer,
 	/* Write the number of items */
 	p11_rpc_buffer_add_uint32 (buffer, count);
 
-	for (i = 0; i < count; i++) {
-		if (mechs[i] > UINT64_MAX) {
-			p11_buffer_fail (buffer);
-			return;
+	if (value) {
+		const CK_MECHANISM_TYPE *mechs = value;
+		size_t i;
+
+		for (i = 0; i < count; i++) {
+			if (mechs[i] > UINT64_MAX) {
+				p11_buffer_fail (buffer);
+				return;
+			}
+			p11_rpc_buffer_add_uint64 (buffer, mechs[i]);
 		}
-		p11_rpc_buffer_add_uint64 (buffer, mechs[i]);
 	}
 }
 
@@ -982,6 +994,7 @@ p11_rpc_buffer_add_date_value (p11_buffer *buffer,
 {
 	CK_DATE date_value;
 	unsigned char array[8];
+	unsigned char *ptr = NULL;
 
 	/* Check if value is empty or can be converted to CK_DATE. */
 	if (value_length != 0 && value_length != sizeof (CK_DATE)) {
@@ -989,14 +1002,15 @@ p11_rpc_buffer_add_date_value (p11_buffer *buffer,
 		return;
 	}
 
-	if (value_length == sizeof (CK_DATE)) {
+	if (value && value_length == sizeof (CK_DATE)) {
 		memcpy (&date_value, value, value_length);
 		memcpy (array, date_value.year, 4);
 		memcpy (array + 4, date_value.month, 2);
 		memcpy (array + 6, date_value.day, 2);
+		ptr = array;
 	}
 
-	p11_rpc_buffer_add_byte_array (buffer, array, value_length);
+	p11_rpc_buffer_add_byte_array (buffer, ptr, value_length);
 }
 
 void
@@ -1212,7 +1226,7 @@ p11_rpc_buffer_get_attribute (p11_buffer *buffer,
 			      size_t *offset,
 			      CK_ATTRIBUTE *attr)
 {
-	uint32_t type, length;
+	uint32_t type, length, decode_length;
 	unsigned char validity;
 	p11_rpc_attribute_serializer *serializer;
 	p11_rpc_value_type value_type;
@@ -1242,8 +1256,13 @@ p11_rpc_buffer_get_attribute (p11_buffer *buffer,
 	assert (serializer != NULL);
 	if (!serializer->decode (buffer, offset, attr->pValue, &attr->ulValueLen))
 		return false;
-	if (!attr->pValue)
+	if (!attr->pValue) {
+		decode_length = attr->ulValueLen;
 		attr->ulValueLen = length;
+		if (decode_length > length) {
+			return false;
+		}
+	}
 	attr->type = type;
 	return true;
 }
