@@ -255,6 +255,10 @@ void             p11_rpc_message_clear                   (p11_rpc_message *msg);
 void *           p11_rpc_message_alloc_extra             (p11_rpc_message *msg,
                                                           size_t length);
 
+void *           p11_rpc_message_alloc_extra_array       (p11_rpc_message *msg,
+                                                          size_t nmemb,
+                                                          size_t size);
+
 bool             p11_rpc_message_prep                    (p11_rpc_message *msg,
                                                           int call_id,
                                                           p11_rpc_message_type type);
@@ -279,16 +283,14 @@ bool             p11_rpc_message_write_space_string      (p11_rpc_message *msg,
                                                                    CK_ULONG length);
 
 bool             p11_rpc_message_write_byte_buffer       (p11_rpc_message *msg,
-                                                          CK_BYTE_PTR array,
-                                                          CK_ULONG_PTR n_array);
+                                                          CK_ULONG count);
 
 bool             p11_rpc_message_write_byte_array        (p11_rpc_message *msg,
                                                           CK_BYTE_PTR arr,
                                                           CK_ULONG num);
 
 bool             p11_rpc_message_write_ulong_buffer      (p11_rpc_message *msg,
-                                                          CK_ULONG_PTR array,
-                                                          CK_ULONG_PTR n_array);
+                                                          CK_ULONG count);
 
 bool             p11_rpc_message_write_ulong_array       (p11_rpc_message *msg,
                                                           CK_ULONG_PTR arr,
