@@ -68,7 +68,6 @@ proto_read_byte_buffer (p11_rpc_message *msg,
                         CK_ULONG *n_buffer)
 {
 	uint32_t length;
-	uint8_t buffer_is_null;
 
 	assert (msg != NULL);
 	assert (buffer != NULL);
@@ -78,8 +77,6 @@ proto_read_byte_buffer (p11_rpc_message *msg,
 	/* Check that we're supposed to be reading this at this point */
 	assert (!msg->signature || p11_rpc_message_verify_part (msg, "fy"));
 
-	if (!p11_rpc_buffer_get_byte (msg->input, &msg->parsed, &buffer_is_null))
-		return PARSE_ERROR;
 	/* The number of ulongs there's room for on the other end */
 	if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &length))
 		return PARSE_ERROR;
@@ -87,11 +84,11 @@ proto_read_byte_buffer (p11_rpc_message *msg,
 	*n_buffer = length;
 	*buffer = NULL;
 
-	/* If buffer is NULL, the caller just wants the length */
-	if (buffer_is_null)
+	/* If set to zero, then they just want the length */
+	if (length == 0)
 		return CKR_OK;
 
-	*buffer = p11_rpc_message_alloc_extra (msg, length * sizeof (CK_BYTE));
+	*buffer = p11_rpc_message_alloc_extra_array (msg, length, sizeof (CK_BYTE));
 	if (*buffer == NULL)
 		return CKR_DEVICE_MEMORY;
 
@@ -169,7 +166,6 @@ proto_read_ulong_buffer (p11_rpc_message *msg,
                          CK_ULONG *n_buffer)
 {
 	uint32_t length;
-	uint8_t buffer_is_null;
 
 	assert (msg != NULL);
 	assert (buffer != NULL);
@@ -179,8 +175,6 @@ proto_read_ulong_buffer (p11_rpc_message *msg,
 	/* Check that we're supposed to be reading this at this point */
 	assert (!msg->signature || p11_rpc_message_verify_part (msg, "fu"));
 
-	if (!p11_rpc_buffer_get_byte (msg->input, &msg->parsed, &buffer_is_null))
-		return PARSE_ERROR;
 	/* The number of ulongs there's room for on the other end */
 	if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &length))
 		return PARSE_ERROR;
@@ -188,11 +182,11 @@ proto_read_ulong_buffer (p11_rpc_message *msg,
 	*n_buffer = length;
 	*buffer = NULL;
 
-	/* If buffer is NULL, the caller just wants the length */
-	if (buffer_is_null)
+	/* If set to zero, then they just want the length */
+	if (length == 0)
 		return CKR_OK;
 
-	*buffer = p11_rpc_message_alloc_extra (msg, length * sizeof (CK_ULONG));
+	*buffer = p11_rpc_message_alloc_extra_array (msg, length, sizeof (CK_ULONG));
 	if (!*buffer)
 		return CKR_DEVICE_MEMORY;
 
@@ -237,7 +231,6 @@ proto_read_attribute_buffer (p11_rpc_message *msg,
 {
 	CK_ATTRIBUTE_PTR attrs;
 	uint32_t n_attrs, i;
-	uint8_t attrs_is_null;
 	uint32_t value;
 
 	assert (msg != NULL);
@@ -252,24 +245,13 @@ proto_read_attribute_buffer (p11_rpc_message *msg,
 	if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &n_attrs))
 		return PARSE_ERROR;
 
-	/* Is attrs pointer null or not ? */
-	if (!p11_rpc_buffer_get_byte (msg->input, &msg->parsed, &attrs_is_null))
-		return PARSE_ERROR;
-
-	if (attrs_is_null) {
-		*result = NULL;
-		*n_result = n_attrs;
-		return CKR_OK;
-	}
-
 	/* Allocate memory for the attribute structures */
-	attrs = p11_rpc_message_alloc_extra (msg, n_attrs * sizeof (CK_ATTRIBUTE));
+	attrs = p11_rpc_message_alloc_extra_array (msg, n_attrs, sizeof (CK_ATTRIBUTE));
 	if (attrs == NULL)
 		return CKR_DEVICE_MEMORY;
 
 	/* Now go through and fill in each one */
 	for (i = 0; i < n_attrs; ++i) {
-		uint8_t value_is_null;
 
 		/* The attribute type */
 		if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &value))
@@ -277,22 +259,18 @@ proto_read_attribute_buffer (p11_rpc_message *msg,
 
 		attrs[i].type = value;
 
-		/* Is the value buffer not NULL ?  */
-		if (!p11_rpc_buffer_get_byte (msg->input, &msg->parsed, &value_is_null))
-			return PARSE_ERROR;
-
 		/* The number of bytes to allocate */
 		if (!p11_rpc_buffer_get_uint32 (msg->input, &msg->parsed, &value))
 			return PARSE_ERROR;
 
-		attrs[i].ulValueLen = value;
-
-		if (value_is_null) {
+		if (value == 0) {
 			attrs[i].pValue = NULL;
+			attrs[i].ulValueLen = 0;
 		} else {
 			attrs[i].pValue = p11_rpc_message_alloc_extra (msg, value);
 			if (!attrs[i].pValue)
 				return CKR_DEVICE_MEMORY;
+			attrs[i].ulValueLen = value;
 		}
 	}
 
@@ -322,7 +300,7 @@ proto_read_attribute_array (p11_rpc_message *msg,
 		return PARSE_ERROR;
 
 	/* Allocate memory for the attribute structures */
-	attrs = p11_rpc_message_alloc_extra (msg, n_attrs * sizeof (CK_ATTRIBUTE));
+	attrs = p11_rpc_message_alloc_extra_array (msg, n_attrs, sizeof (CK_ATTRIBUTE));
 	if (attrs == NULL)
 		return CKR_DEVICE_MEMORY;
 
