@@ -56,6 +56,14 @@ is_module_empty (P11KitUri *uri)
 	        info->libraryVersion.minor == (CK_BYTE)-1);
 }
 
+static int
+is_slot_empty (P11KitUri *uri)
+{
+	CK_SLOT_INFO_PTR slot = p11_kit_uri_get_slot_info (uri);
+	return (slot->slotDescription[0] == 0 &&
+	        slot->manufacturerID[0] == 0);
+}
+
 static int
 is_token_empty (P11KitUri *uri)
 {
@@ -87,6 +95,27 @@ test_uri_parse (void)
 	assert_num_eq (P11_KIT_URI_OK, ret);
 
 	assert (is_module_empty (uri));
+	assert (is_slot_empty (uri));
+	assert (is_token_empty (uri));
+	assert (are_attributes_empty (uri));
+
+	p11_kit_uri_free (uri);
+}
+
+static void
+test_uri_parse_case_insensitive (void)
+{
+	P11KitUri *uri;
+	int ret;
+
+	uri = p11_kit_uri_new ();
+	assert_ptr_not_null (uri);
+
+	ret = p11_kit_uri_parse ("PKCS11:", P11_KIT_URI_FOR_MODULE, uri);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+
+	assert (is_module_empty (uri));
+	assert (is_slot_empty (uri));
 	assert (is_token_empty (uri));
 	assert (are_attributes_empty (uri));
 
@@ -122,6 +151,7 @@ test_uri_parse_with_label (void)
 	assert_num_eq (P11_KIT_URI_OK, ret);
 
 	assert (is_module_empty (uri));
+	assert (is_slot_empty (uri));
 	assert (is_token_empty (uri));
 
 	attr = p11_kit_uri_get_attribute (uri, CKA_LABEL);
@@ -428,6 +458,27 @@ test_uri_parse_with_library_bad_encoding (void)
 	p11_kit_uri_free (uri);
 }
 
+static void
+test_uri_parse_with_slot (void)
+{
+	P11KitUri *uri = NULL;
+	CK_SLOT_INFO_PTR slot;
+	int ret;
+
+	uri = p11_kit_uri_new ();
+	assert_ptr_not_null (uri);
+
+	ret = p11_kit_uri_parse ("pkcs11:slot-description=Slot%20Description;slot-manufacturer=Me",
+	                         P11_KIT_URI_FOR_SLOT, uri);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+
+	slot = p11_kit_uri_get_slot_info (uri);
+	assert (is_space_string (slot->slotDescription, sizeof (slot->slotDescription), "Slot Description"));
+	assert (is_space_string (slot->manufacturerID, sizeof (slot->manufacturerID), "Me"));
+
+	p11_kit_uri_free (uri);
+}
+
 static void
 test_uri_build_empty (void)
 {
@@ -606,12 +657,49 @@ test_uri_build_with_attributes (void)
 
 	assert (strstr (string, "object=The%20Label") != NULL);
 	assert (strstr (string, "type=data") != NULL);
-	assert (strstr (string, "id=%48%45%4c%4c%4f") != NULL);
+	assert (strstr (string, "id=%48%45%4C%4C%4F") != NULL);
 
 	free (string);
 	p11_kit_uri_free (uri);
 }
 
+static void
+test_uri_build_with_slot_info (void)
+{
+	char *string = NULL;
+	P11KitUri *uri;
+	P11KitUri *check;
+	CK_SLOT_INFO_PTR slot;
+	int ret;
+
+	uri = p11_kit_uri_new ();
+	assert_ptr_not_null (uri);
+
+	slot = p11_kit_uri_get_slot_info (uri);
+	set_space_string (slot->slotDescription, sizeof (slot->slotDescription), "The Slot Description");
+	set_space_string (slot->manufacturerID, sizeof (slot->manufacturerID), "Me");
+
+	ret = p11_kit_uri_format (uri, P11_KIT_URI_FOR_ANY, &string);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+	assert_ptr_not_null (string);
+
+	check = p11_kit_uri_new ();
+	assert_ptr_not_null (check);
+
+	ret = p11_kit_uri_parse (string, P11_KIT_URI_FOR_SLOT, check);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+
+	p11_kit_uri_match_slot_info (check, p11_kit_uri_get_slot_info (uri));
+
+	p11_kit_uri_free (uri);
+	p11_kit_uri_free (check);
+
+	assert (strstr (string, "slot-description=The%20Slot%20Description") != NULL);
+	assert (strstr (string, "slot-manufacturer=Me") != NULL);
+
+	free (string);
+}
+
 static void
 test_uri_parse_private_key (void)
 {
@@ -1264,10 +1352,10 @@ test_uri_pin_source (void)
 
 	ret = p11_kit_uri_format (uri, P11_KIT_URI_FOR_ANY, &string);
 	assert_num_eq (P11_KIT_URI_OK, ret);
-	assert (strstr (string, "pin-source=%7cmy-pin-file") != NULL);
+	assert (strstr (string, "pin-source=%7Cmy-pin-file") != NULL);
 	free (string);
 
-	ret = p11_kit_uri_parse ("pkcs11:pin-source=blah%2Fblah", P11_KIT_URI_FOR_ANY, uri);
+	ret = p11_kit_uri_parse ("pkcs11:?pin-source=blah%2Fblah", P11_KIT_URI_FOR_ANY, uri);
 	assert_num_eq (P11_KIT_URI_OK, ret);
 
 	pin_source = p11_kit_uri_get_pin_source (uri);
@@ -1300,10 +1388,10 @@ test_uri_pin_value (void)
 
 	ret = p11_kit_uri_format (uri, P11_KIT_URI_FOR_ANY, &string);
 	assert_num_eq (P11_KIT_URI_OK, ret);
-	assert (strstr (string, "pkcs11:pin-value=1%2a%26%23%25%26%40%28") != NULL);
+	assert (strstr (string, "pkcs11:?pin-value=1%2A%26%23%25%26%40%28") != NULL);
 	free (string);
 
-	ret = p11_kit_uri_parse ("pkcs11:pin-value=blah%2Fblah", P11_KIT_URI_FOR_ANY, uri);
+	ret = p11_kit_uri_parse ("pkcs11:?pin-value=blah%2Fblah", P11_KIT_URI_FOR_ANY, uri);
 	assert_num_eq (P11_KIT_URI_OK, ret);
 
 	pin_value = p11_kit_uri_get_pin_value (uri);
@@ -1321,12 +1409,217 @@ test_uri_pin_value_bad (void)
 	uri = p11_kit_uri_new ();
 	assert_ptr_not_null (uri);
 
-	ret = p11_kit_uri_parse ("pkcs11:pin-value=blahblah%2", P11_KIT_URI_FOR_ANY, uri);
+	ret = p11_kit_uri_parse ("pkcs11:?pin-value=blahblah%2", P11_KIT_URI_FOR_ANY, uri);
 	assert_num_eq (P11_KIT_URI_BAD_ENCODING, ret);
 
 	p11_kit_uri_free (uri);
 }
 
+static void
+test_uri_module_name (void)
+{
+	P11KitUri *uri;
+	const char *module_name;
+	char *string;
+	int ret;
+
+	uri = p11_kit_uri_new ();
+	assert_ptr_not_null (uri);
+
+	p11_kit_uri_set_module_name (uri, "123456");
+
+	module_name = p11_kit_uri_get_module_name (uri);
+	assert_str_eq ("123456", module_name);
+
+	p11_kit_uri_set_module_name (uri, "1*&#%&@(");
+
+	module_name = p11_kit_uri_get_module_name (uri);
+	assert_str_eq ("1*&#%&@(", module_name);
+
+	ret = p11_kit_uri_format (uri, P11_KIT_URI_FOR_ANY, &string);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+	assert (strstr (string, "pkcs11:?module-name=1%2A%26%23%25%26%40%28") != NULL);
+	free (string);
+
+	ret = p11_kit_uri_parse ("pkcs11:?module-name=blah%2Fblah", P11_KIT_URI_FOR_ANY, uri);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+
+	module_name = p11_kit_uri_get_module_name (uri);
+	assert_str_eq ("blah/blah", module_name);
+
+	p11_kit_uri_free (uri);
+}
+
+static void
+test_uri_module_name_bad (void)
+{
+	P11KitUri *uri;
+	int ret;
+
+	uri = p11_kit_uri_new ();
+	assert_ptr_not_null (uri);
+
+	ret = p11_kit_uri_parse ("pkcs11:?module-name=blahblah%2", P11_KIT_URI_FOR_ANY, uri);
+	assert_num_eq (P11_KIT_URI_BAD_ENCODING, ret);
+
+	p11_kit_uri_free (uri);
+}
+
+static void
+test_uri_module_path (void)
+{
+	P11KitUri *uri;
+	const char *module_path;
+	char *string;
+	int ret;
+
+	uri = p11_kit_uri_new ();
+	assert_ptr_not_null (uri);
+
+	p11_kit_uri_set_module_path (uri, "/my-module-path");
+
+	module_path = p11_kit_uri_get_module_path (uri);
+	assert_str_eq ("/my-module-path", module_path);
+
+	ret = p11_kit_uri_format (uri, P11_KIT_URI_FOR_ANY, &string);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+	assert (strstr (string, "module-path=%2Fmy-module-path") != NULL);
+	free (string);
+
+	ret = p11_kit_uri_parse ("pkcs11:?module-path=blah%2Fblah", P11_KIT_URI_FOR_ANY, uri);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+
+	module_path = p11_kit_uri_get_module_path (uri);
+	assert_str_eq ("blah/blah", module_path);
+
+	p11_kit_uri_free (uri);
+}
+
+static void
+test_uri_module_name_and_path (void)
+{
+	P11KitUri *uri;
+	const char *module_name;
+	const char *module_path;
+	char *string;
+	int ret;
+
+	uri = p11_kit_uri_new ();
+	assert_ptr_not_null (uri);
+
+	p11_kit_uri_set_module_name (uri, "123456");
+	p11_kit_uri_set_module_path (uri, "/my-module-path");
+
+	ret = p11_kit_uri_format (uri, P11_KIT_URI_FOR_ANY, &string);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+	assert (strstr (string, "pkcs11:?module-name=123456&module-path=%2Fmy-module-path") != NULL);
+	free (string);
+
+	ret = p11_kit_uri_parse ("pkcs11:?module-name=1%2A%26%23%25%26%40%28&module-path=blah%2Fblah", P11_KIT_URI_FOR_ANY, uri);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+
+	module_name = p11_kit_uri_get_module_name (uri);
+	assert_str_eq ("1*&#%&@(", module_name);
+	module_path = p11_kit_uri_get_module_path (uri);
+	assert_str_eq ("blah/blah", module_path);
+
+	p11_kit_uri_free (uri);
+}
+
+static void
+test_uri_vendor_query (void)
+{
+	P11KitUri *uri;
+	const char *value;
+	char *string;
+	int ret;
+
+	uri = p11_kit_uri_new ();
+	assert_ptr_not_null (uri);
+
+	ret = p11_kit_uri_set_vendor_query (uri, "my-query-one", "123456");
+	assert_num_eq (1, ret);
+	value = p11_kit_uri_get_vendor_query (uri, "my-query-one");
+	assert_str_eq ("123456", value);
+
+	ret = p11_kit_uri_format (uri, P11_KIT_URI_FOR_ANY, &string);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+	assert (strstr (string, "my-query-one=123456") != NULL);
+	free (string);
+
+	ret = p11_kit_uri_parse ("pkcs11:?my-query-two=some-value", P11_KIT_URI_FOR_ANY, uri);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+
+	value = p11_kit_uri_get_vendor_query (uri, "my-query-two");
+	assert_str_eq ("some-value", value);
+
+	ret = p11_kit_uri_set_vendor_query (uri, "my-query-two", "other-value");
+	assert_num_eq (1, ret);
+
+	value = p11_kit_uri_get_vendor_query (uri, "my-query-two");
+	assert_str_eq ("other-value", value);
+
+	ret = p11_kit_uri_set_vendor_query (uri, "my-query-three", NULL);
+	assert_num_eq (0, ret);
+
+	/* Check if duplicate vendor query attributes are accepted and
+	 * sorted alphabetically.  */
+	ret = p11_kit_uri_parse ("pkcs11:?bbb=zzz&aaa=xxx&aaa=yyy", P11_KIT_URI_FOR_ANY, uri);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+
+	ret = p11_kit_uri_format (uri, P11_KIT_URI_FOR_ANY, &string);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+	assert_str_eq ("pkcs11:?aaa=xxx&aaa=yyy&bbb=zzz", string);
+	free (string);
+
+	p11_kit_uri_free (uri);
+}
+
+static void
+test_uri_slot_id (void)
+{
+	P11KitUri *uri;
+	CK_SLOT_ID slot_id;
+	char *string;
+	int ret;
+
+	uri = p11_kit_uri_new ();
+	assert_ptr_not_null (uri);
+
+	p11_kit_uri_set_slot_id (uri, 12345);
+
+	slot_id = p11_kit_uri_get_slot_id (uri);
+	assert_num_eq (12345, slot_id);
+
+	ret = p11_kit_uri_format (uri, P11_KIT_URI_FOR_ANY, &string);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+	assert (strstr (string, "pkcs11:slot-id=12345") != NULL);
+	free (string);
+
+	ret = p11_kit_uri_parse ("pkcs11:slot-id=67890", P11_KIT_URI_FOR_ANY, uri);
+	assert_num_eq (P11_KIT_URI_OK, ret);
+
+	slot_id = p11_kit_uri_get_slot_id (uri);
+	assert_num_eq (67890, slot_id);
+
+	p11_kit_uri_free (uri);
+}
+
+static void
+test_uri_slot_id_bad (void)
+{
+	P11KitUri *uri;
+	int ret;
+
+	uri = p11_kit_uri_new ();
+	assert_ptr_not_null (uri);
+
+	ret = p11_kit_uri_parse ("pkcs11:slot-id=123^456", P11_KIT_URI_FOR_ANY, uri);
+	assert_num_eq (P11_KIT_URI_BAD_SYNTAX, ret);
+
+	p11_kit_uri_free (uri);
+}
+
 static void
 test_uri_free_null (void)
 {
@@ -1346,6 +1639,7 @@ main (int argc,
       char *argv[])
 {
 	p11_test (test_uri_parse, "/uri/test_uri_parse");
+	p11_test (test_uri_parse_case_insensitive, "/uri/test_uri_parse_case_insensitive");
 	p11_test (test_uri_parse_bad_scheme, "/uri/test_uri_parse_bad_scheme");
 	p11_test (test_uri_parse_with_label, "/uri/test_uri_parse_with_label");
 	p11_test (test_uri_parse_with_empty_label, "/uri/test_uri_parse_with_empty_label");
@@ -1361,11 +1655,13 @@ main (int argc,
 	p11_test (test_uri_parse_with_spaces, "/uri/test_uri_parse_with_spaces");
 	p11_test (test_uri_parse_with_library, "/uri/test_uri_parse_with_library");
 	p11_test (test_uri_parse_with_library_bad_encoding, "/uri/test_uri_parse_with_library_bad_encoding");
+	p11_test (test_uri_parse_with_slot, "/uri/test_uri_parse_with_slot");
 	p11_test (test_uri_build_empty, "/uri/test_uri_build_empty");
 	p11_test (test_uri_build_with_token_info, "/uri/test_uri_build_with_token_info");
 	p11_test (test_uri_build_with_token_null_info, "/uri/test_uri_build_with_token_null_info");
 	p11_test (test_uri_build_with_token_empty_info, "/uri/test_uri_build_with_token_empty_info");
 	p11_test (test_uri_build_with_attributes, "/uri/test_uri_build_with_attributes");
+	p11_test (test_uri_build_with_slot_info, "/uri/test_uri_build_with_slot_info");
 	p11_test (test_uri_parse_private_key, "/uri/test_uri_parse_private_key");
 	p11_test (test_uri_parse_secret_key, "/uri/test_uri_parse_secret_key");
 	p11_test (test_uri_parse_library_version, "/uri/test_uri_parse_library_version");
@@ -1388,6 +1684,13 @@ main (int argc,
 	p11_test (test_uri_pin_source, "/uri/test_uri_pin_source");
 	p11_test (test_uri_pin_value, "/uri/pin-value");
 	p11_test (test_uri_pin_value_bad, "/uri/pin-value-bad");
+	p11_test (test_uri_module_name, "/uri/module-name");
+	p11_test (test_uri_module_name_bad, "/uri/module-name-bad");
+	p11_test (test_uri_module_path, "/uri/module-path");
+	p11_test (test_uri_module_name_and_path, "/uri/module-name-and-path");
+	p11_test (test_uri_vendor_query, "/uri/vendor-query");
+	p11_test (test_uri_slot_id, "/uri/slot-id");
+	p11_test (test_uri_slot_id_bad, "/uri/slot-id-bad");
 	p11_test (test_uri_free_null, "/uri/test_uri_free_null");
 	p11_test (test_uri_message, "/uri/test_uri_message");
 
