@@ -77,6 +77,7 @@ setup (void *unused)
 	assert_num_eq (CKR_OK, rv);
 
 	p11_enumerate_init (&test.ex);
+	test.ex.flags |= P11_ENUMERATE_CORRELATE;
 
 	test.directory = p11_test_directory ("test-extract");
 }
@@ -594,11 +595,14 @@ test_directory (void)
 	ret = p11_extract_openssl_directory (&test.ex, test.directory);
 	assert_num_eq (true, ret);
 
-	test_check_directory (test.directory, ("Custom_Label.pem", "Custom_Label.1.pem",
 #ifdef OS_UNIX
-	                                           "e5662767.1", "e5662767.0", "590d426f.1", "590d426f.0",
+	test_check_directory (test.directory, ("Custom_Label.pem", "Custom_Label.1.pem",
+					       "e5662767.1", "e5662767.0", "590d426f.1", "590d426f.0",
+					       NULL));
+#else
+	test_check_directory (test.directory, ("Custom_Label.pem", "Custom_Label.1.pem",
+					       NULL));
 #endif
-	                                           NULL));
 	test_check_file (test.directory, "Custom_Label.pem",
 	                 SRCDIR "/trust/fixtures/cacert3-trusted-server-alias.pem");
 	test_check_file (test.directory, "Custom_Label.1.pem",
